file=Makefile
echo "Extracting $file"
rm -f $file
tarfile=pari-$revision.tar

file=Makefile.$osname-$arch$suffix
ln -s $file Makefile 
rm -f $file

cat > $file << EOT
#
# This file was created by Configure. All changes made will be lost
# next time Configure is run.
#
dft target::
	@echo "Possible targets are :"
	@echo "	gp			Compilation"
	@echo "	bench, test 		Compilation and test"
	@echo "	doc			Documentation"
	@echo "	install			Installation"
	@echo "For pari maintainers..."
	@echo "	all			Compile gp-sta gp-dyn lib-sta lib-dyn"
	@echo "	distrib			For distribution of pari/gp"
	@echo "	bigdistrib		Idem, with all directories"
	@echo "	exp			Copy of the sources for experimental release"
	@echo "	rel			New release"
	@echo "	patch			Experimental diff file"
	@echo "	undo			Undo experimental release"

gp all ftest bench test test-compat install clean::
	cd $objdir; \$(MAKE) \$@

doc::
	cd doc; \$(MAKE)

undo::
	@for dir in configure src doc perl; do /bin/rm -rf \$\$dir; \\
	   mv \$\$dir.orig \$\$dir; done
	@rm exp

exp::
	@for dir in configure src doc perl; do if test -d \$\$dir.orig; then \\
	   echo "Remove \$\$dir.orig before building an experimental release";\\
	   exit 1; fi; done
	for dir in configure src doc perl; do \\
	   mv \$\$dir \$\$dir.orig; cp -pr \$\$dir.orig \$\$dir; done
	ln -s src exp
	sed "s,patch_level=.*,patch_level=exp," \\
	    < configure.orig/version > configure/version

patch:: pl$next_patch
pl$next_patch:
	@if diff -c src exp >/dev/null 2>/dev/null; then :; else \\
	   echo "'src' should be a link to 'exp'"; exit 1; fi
	-diff -c Configure.orig Configure > pl$next_patch
	@for dir in configure src doc perl; do \\
	   echo "diff -c \$\$dir.orig \$\$dir"; \\
	   for i in \`find \$\$dir -type f -print\`; do \\
	   diff -c \`echo \$\$i|sed s,\$\$dir/,\$\$dir.orig/,\` \$\$i; \\
	   done >> pl$next_patch; done

update-version::
	sed -e "s,^patch_level=.*,patch_level=$next_patch," \\
	    -e "s,^next_patch=.*,next_patch=`expr $next_patch + 1`," \\
	  < configure.orig/version > configure/version

rel: update-version pl$next_patch
	rm -f exp

RCS::
	cd src; ci -l -r\$\${REVISION:-$revision} -m"." */*.[chs] test/in

setdir::
	@dir=pari-$revision; if test -d \$\$dir; then \\
	   echo "Remove \$\$dir before building a new release"; exit 1;\\
	fi;\\
	mkdir \$\$dir; mv $tarfile \$\$dir; \\
	cd \$\$dir; tar xf $tarfile; rm -f $tarfile; cd ..; \\
	tar cf $tarfile \$\$dir

bigdistrib: bigtar setdir
	gzip $tarfile

distrib: tar setdir
	gzip $tarfile

bigtar: tar
	tar rf $tarfile doc/gp.1 doc/t* doc/app* doc/usersch* doc/*.in doc/M*

tar::
EOT
case "`tar 2>&1`" in
  *"--help"*) cat >> $file << EOT
	tar cfXT $tarfile $config_dir/tar_exclude $config_dir/tar_include
EOT
;;
  *"-I include-file"*) cat >> $file << EOT
	tar FcfX $tarfile $config_dir/tar_exclude -I $config_dir/tar_include
EOT
;;
esac
