/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                     OPERATIONS GENERIQUES                      **/
/**                       (premiere partie)                        **/
/**                                                                **/
/**                      copyright Babe Cool                       **/
/**                                                                **/
/********************************************************************/
/********************************************************************/
/* $Id: gen1.c,v 1.922 1997/04/06 02:52:40 karim Exp karim $ */

# include "genpari.h"

/***********************************************************************/
/*                    LISTE DES TYPES GENERIQUES                       */
/*                                                                     */
/*  t_INT    : entier long   [ cod1 ] [ cod2 ] [ man1 ] ... [ manl ]   */
/*  t_REAL   : reel          [ cod1 ] [ cod2 ] [ man1 ] ... [ manl ]   */
/*  t_INTMOD : entier modulo [ code ] [ mod  ] [ entier modulo ]       */
/*  t_FRAC   : fraction      [ code ] [ num. ] [ den. ]                */
/*  t_FRACN  : nfraction     [ code ] [ num. ] [ den. ]                */
/*  t_COMPLEX: complexe      [ code ] [ reel ] [ imag ]                */
/*  t_PADIC  : p-adique      [ cod1 ] [ cod2 ] [ p ] [ p^r ] [ entier] */
/*  t_QUAD   : quadrat       [ cod1 ] [ mod  ] [ reel ] [ imag ]       */
/*  t_POLMOD : poly mod      [ code ] [ mod  ] [ polynome  mod ]       */
/* ------------------------------------------------------------------  */
/*  t_POL    : polynome      [ cod1 ] [ cod2 ] [ man1 ] ... [ manl ]   */
/*  t_SER    : serie         [ cod1 ] [ cod2 ] [ man1 ] ... [ manl ]   */
/*  t_RFRAC  : fr.rat        [ code ] [ num. ] [ den. ]                */
/*  t_RFRACN : n.fr.rat      [ code ] [ num. ] [ den. ]                */
/*  t_QFR    : formqre       [ code ] [  a  ] [  b  ] [  c  ] [ del ]  */
/*  t_QFI    : formqim       [ code ] [  a   ] [  b   ] [  c   ]       */
/*  t_VEC    : vect. ligne   [ code ] [  x1  ] ... [  xl  ]            */
/*  t_COL    : vect. colonne [ code ] [  x1  ] ... [  xl  ]            */
/*  t_MAT    : matrice       [ code ] [ col1 ] ... [ coll ]            */
/*                                                                     */
/***********************************************************************/

/********************************************************************/
/**                                                                **/
/**                       ADDITION GENERALE                        **/
/**                                                                **/
/********************************************************************/

/* addition d'un type entier ou rationnel avec un p-adique
 * x doit etre entier ou rationnel et y p-adique. A usage interne
 */
static GEN
gaddpex(GEN x, GEN y)
{
  GEN z,p,p1,p2,p3;
  long tx,e1,e2,e3,av,tetpil;

  if (gcmp0(x)) return gcopy(y);

  av=avma; z=cgetg(5,t_PADIC); e1=valp(y);
  p=(GEN)y[2]; tx=typ(x);
  if (is_frac_t(tx))
  {
    e3 = pvaluation((GEN)x[1],p,&p2) - pvaluation((GEN)x[2],p,&p3);
    p2=gdiv(p2,p3);
  }
  else e3=pvaluation(x,p,&p2);
  e1=e1-e3; e2=signe(y[4]) ? e1+precp(y) : e1;
  z[2]=lcopy(p);
  if (e2<=0) { z[3]=un; z[1]=evalprecp(0)+evalvalp(e3); }
  else
  {
    z[1]=evalprecp(e2)+evalvalp(e3);
    if (e1)
    {
      p1=gpuigs(p,e1);
      z[3]=lmul((GEN)y[3],p1);
    }
    else z[3]=lcopy((GEN)y[3]);
  }
  z[4]=lmod(p2,(GEN)z[3]);
  tetpil=avma; return gerepile(av,tetpil,gadd(z,y));
}

/*  a usage interne: aucune verification de types  */
static long
kro8(GEN x, GEN y)
{
  GEN  p1,p2;
  long k,av=avma;

  p1=(GEN)x[1]; p2=(GEN)(p1[3]);
  p1=gsub(gsqr(p2), gmul2n((GEN)p1[2],2));
  k=kronecker(p1,y); avma=av; return k;
}

GEN
gadd(GEN x, GEN y)
{
  long vx,vy,lx,ly,lz,tx,ty,i,j,k,l;
  long av,tetpil;
  GEN z,p1,p2;

  tx=typ(x); ty=typ(y);
  if (is_const_t(tx) && is_const_t(ty))
  {
    if (tx>ty) { p1=x; x=y; y=p1; i=tx; tx=ty; ty=i; }
  }
  else
  {
    vx=gvar(x); vy=gvar(y);
    if (vx<vy || (vx==vy && tx>ty))
    {
      p1=x; x=y; y=p1;
      i=tx; tx=ty; ty=i;
      i=vx; vx=vy; vy=i;
    }
  }
  lx=lg(x); ly=lg(y);

  if (is_scalar_t(ty))
  {
    switch(tx)
    {
      case t_INT:
        switch(ty)
	{
	  case t_INT: case t_REAL:
	    return mpadd(x,y);
	  
	  case t_INTMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    l=avma; p1=addii(x,(GEN)y[2]);
	    tetpil=avma; p1=modii(p1,(GEN)y[1]);
	    z[2]=lpile(l,tetpil,p1);
	    return z;

	  case t_FRAC: case t_FRACN:
	    z=cgetg(ly,ty); l=avma;
	    p1=mulii((GEN)y[2],x);
	    tetpil=avma; p1=addii(p1,(GEN)y[1]);
	    z[1]=lpile(l,tetpil,p1);
	    z[2]=lcopy((GEN)y[2]);
	    return z;

	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=ladd(x,(GEN)y[1]);
	    z[2]=lcopy((GEN)y[2]);
	    return z;
	  
	  case t_PADIC:
	    return gaddpex(x,y);
	  
	  case t_QUAD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=ladd(x,(GEN)y[2]);
	    z[3]=lcopy((GEN)y[3]);
	    return z;
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty); 
	    z[1]=copyifstack(y[1]);
	    z[2]=ladd(x,(GEN)y[2]);
	    return z;
	  default: err(typeer,"addition");
	}
	
      case t_REAL:
        switch(ty)
	{
	  case t_REAL:
	    return mpadd(x,y);
	  
	  case t_FRAC: case t_FRACN:
	    if (signe(y[1]))
	    {
	      if (gcmp0(x))
	      {
		lx=(expi((GEN)y[1])-expi((GEN)y[2])-expo(x))>>TWOPOTBITS_IN_LONG;
		if (lx<0) lx=0;
		lx+=3; z=cgetr(lx); diviiz((GEN)y[1],(GEN)y[2],z);
		return z;
	      }

	      l=avma; z=addir((GEN)y[1],mulir((GEN)y[2],x)); tetpil=avma;
	      return gerepile(l,tetpil,divri(z,(GEN)y[2]));
	    }
	    return gcopy(x);
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=ladd(x,(GEN)y[1]);
	    z[2]=lcopy((GEN)y[2]);
	    return z;

	  case t_QUAD:
	    if (gcmp0(y)) return gcopy(x);

	    l=avma; i=gexpo(y)-expo(x);
	    if (i<0) i=0;
	    p1=co8(y,lx+(i>>TWOPOTBITS_IN_LONG)); tetpil=avma;
	    return gerepile(l,tetpil,gadd(p1,x));
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=ladd(x,(GEN)y[2]);
	    return z;
	  
	  case t_INTMOD: case t_PADIC:
	    err(gadder2);
	  
	  default: err(typeer,"addition");
	}
	
      case t_INTMOD:
        switch(ty)
	{
	  case t_INTMOD:
	    z=cgetg(ly,ty); k=x[1]; l=y[1];
	    if (k==l || gegal((GEN)k,(GEN)l))
	      z[1]=copyifstack(k);
	    else
	      z[1]=lmppgcd((GEN)k,(GEN)l);
	    l=avma; p1=addii((GEN)x[2],(GEN)y[2]); tetpil=avma;
	    z[2]=lpile(l,tetpil,modii(p1,(GEN)z[1]));
	    return z;
	  
	  case t_FRAC: case t_FRACN:
	    z=cgetg(3,t_INTMOD);
	    z[1]=copyifstack(x[1]);
	    z[2]=lgeti(lgefint(x[1]));
	    gaffect(y,z); gaddz(z,x,z);
	    return z;
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty); z[2]=lcopy((GEN)y[2]);
	    z[1]=ladd(x,(GEN)y[1]); return z;
	  
	  case t_PADIC:
	    l=avma; p1=cgetg(3,t_INTMOD);
	    p1[1]=x[1]; p1[2]=lgeti(lgefint(x[1]));
	    gaffect(y,p1); tetpil=avma;
	    return gerepile(l,tetpil,gadd(p1,x));
	  
	  case t_QUAD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]); z[3]=lcopy((GEN)y[3]);
	    z[2]=ladd(x,(GEN)y[2]); return z;
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty); z[2]=ladd(x,(GEN)y[2]);
	    z[1]=copyifstack(y[1]); return z;
	  
	  default: err(typeer,"addition");
	}
	
      case t_FRAC: case t_FRACN:
        switch (ty)
	{
	  case t_FRAC: case t_FRACN:
	    if (tx == t_FRAC && ty == t_FRAC) i=t_FRAC; else i=t_FRACN;
	    z=cgetg(ly,i); l=avma;
	    p1=mulii((GEN)x[1],(GEN)y[2]);
	    p2=mulii((GEN)x[2],(GEN)y[1]);
	    tetpil=avma; z[1]=lpile(l,tetpil,addii(p1,p2));
	    z[2]=lmulii((GEN)x[2],(GEN)y[2]);
	    if (i==t_FRAC) gredsp(&z);
	    return z;

	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=ladd((GEN)y[1],x);
	    z[2]=lcopy((GEN)y[2]);
	    return z;

	  case t_PADIC:
	    return gaddpex(x,y);

	  case t_QUAD:
	    z=cgetg(ly,ty);
	    z[1]=lcopy((GEN)y[1]);
	    z[2]=ladd((GEN)y[2],x);
	    z[3]=lcopy((GEN)y[3]);
	    return z;

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]); z[2]=ladd(x,(GEN)y[2]);
	    return z;
	  
	  default: err(typeer,"addition");
	}
	
      case t_COMPLEX:
        switch(ty)
	{
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=ladd((GEN)x[1],(GEN)y[1]);
	    z[2]=ladd((GEN)x[2],(GEN)y[2]);
	    return z;

	  case t_PADIC:
	    if (krosg(-1,(GEN)y[2])== -1)
	    {
	      z=cgetg(3,t_COMPLEX); z[1]=ladd((GEN)x[1],y);
	      z[2]=lcopy((GEN)x[2]); return z;
	    }
	    l=avma;
	    p1=cvtop(x,(GEN)y[2],signe(y[4])?(valp(y)+precp(y)):valp(y)+1);
	    tetpil=avma; return gerepile(l,tetpil,gadd(p1,y));

	  case t_QUAD:
	    lx=precision(x); if (!lx) err(gadder12);
	    if (gcmp0(y)) return gcopy(x);

	    l=avma; i=gexpo(y)-gexpo(x);
	    if (i<0) i=0;
	    p1=co8(y,lx+(i>>TWOPOTBITS_IN_LONG)); tetpil=avma;
	    return gerepile(l,tetpil,gadd(p1,x));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]); z[2]=ladd(x,(GEN)y[2]);
	    return z;
	  
	  default: err(gadder12);
	}
	
      case t_PADIC:
        switch(ty)
	{
	  case t_PADIC:
	  {
	    long e,r,d,a1,a2,r1,r2,co;
	    GEN p;

	    p=(GEN)x[2];
	    if (cmpii(p,(GEN)y[2])) err(gadder15);
	    a1=valp(x); a2=valp(y); e=a1; d=a2-a1;
	    if (a2<a1)
	    {
	      z=x; x=y; y=z; e=a2; d= -d;
	    }
	    r1=precp(x); r2=precp(y);
	    if (d)
	    {
	      l=avma; p1=gpuigs(p,d); tetpil=avma;
	      z=cgetg(5,t_PADIC); z[2]=(long)copyifstack(p); r=d+r2;
	      if (r1<=r)
	      {
		r=r1; z[3]=lcopy((GEN)x[3]);
	      }
	      else z[3]=lmul(p1,(GEN)y[3]);
	      z[4]=lgeti(lgefint(z[3]));
	      av=avma; p1=mulii(p1,(GEN)y[4]);
	      modiiz(addii(p1,(GEN)x[4]),(GEN)z[3],(GEN)z[4]); avma=av;
	      z[1]=evalprecp(r)+evalvalp(e);
	      return gerepile(l,tetpil,z);
	    }

	    if (r2<r1) { r=r2; z=x; x=y; y=z; } else r=r1;
	    l=avma; z=cgetg(5,t_PADIC); z[2]=(long)copyifstack(p); av=avma;
	    p1=addii((GEN)x[4],(GEN)y[4]);
	    if (gcmp0(p1)||((co=pvaluation(p1,p,&p2))>=r))
	    {
	      avma=av; z[1]=evalvalp(e+r); z[3]=un;
	      z[4]=lgeti(lgefint(x[3]));
	      affsi(0,(GEN)z[4]); return z;
	    }

	    if (co)
	    {
	      p1=gpuigs(p,co);
	      z[3]=ldivii((GEN)x[3],p1);
	      r-=co; z[1]=evalprecp(r)+evalvalp(e+co);
	      z[4]=lgeti(lgefint(z[3]));
	      z[4]=(long)modii(p2,(GEN)z[3]); tetpil=avma;
	      return gerepile(l,tetpil,gcopy(z));
	    }

	    tetpil=avma; z[4]=lgeti(lgefint(x[3]));
	    modiiz(p1,(GEN)x[3],(GEN)z[4]);
	    z[4]=lpile(av,tetpil,(GEN)z[4]);
	    z[3]=lcopy((GEN)x[3]);
	    z[1]=evalprecp(r)+evalvalp(e);
	    return z;
	  }  
	  case t_QUAD:
	    if (kro8(y,(GEN)x[2])== -1)
	    {
	      z=cgetg(4,t_QUAD);
	      z[1]=copyifstack(y[1]);
	      z[2]=ladd((GEN)y[2],x);
	      z[3]=lcopy((GEN)y[3]);
	      return z;
	    }
	    l=avma;
	    p1=cvtop(y,(GEN)x[2],signe(x[4])?(valp(x)+precp(x)):valp(x)+1);
	    tetpil=avma;
	    return gerepile(l,tetpil,gadd(p1,x));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]); z[2]=ladd(x,(GEN)y[2]);
	    return z;

	  default: err(typeer,"addition");
	}

      case t_QUAD:
        switch(ty)
	{
	  case t_QUAD:
	    k=x[1]; l=y[1];
	    if (k!=l && !gegal((GEN)k,(GEN)l)) err(gadder13);
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(l);
	    z[2]=ladd((GEN)x[2],(GEN)y[2]);
	    z[3]=ladd((GEN)x[3],(GEN)y[3]);
	    return z;

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]); z[2]=ladd(x,(GEN)y[2]);
	    return z;
	  default: err(typeer,"addition");
	}

      case t_POLMOD:
        z=cgetg(lx,tx); k=x[1]; l=y[1];
	if ((k==l)||gegal((GEN)k,(GEN)l))
	{ 
	  z[1]=copyifstack(k); l=avma;
	  p1=gadd((GEN)x[2],(GEN)y[2]);
	}
	else
	{
	  vx=varn(x[1]); vy=varn(y[1]);
	  if (vx==vy)
	  { 
	    z[1]=lgcd((GEN)k,(GEN)l); l=avma;
	    p1=gadd((GEN)x[2],(GEN)y[2]);
	  }
	  else
	  {
	    if (vx<vy)
	    { z[1]=copyifstack(k); l=avma; p1=gadd((GEN)x[2],y); }
	    else
	    { z[1]=copyifstack(l); l=avma; p1=gadd((GEN)y[2],x); }
	  }
	}
	tetpil=avma;
	z[2]=lpile(l,tetpil,gmod(p1,(GEN)z[1]));
	return z;

      default: if (ty!=t_POLMOD) err(gaddbug1);
	z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	l=avma; p1=gadd(x,(GEN)y[2]); tetpil=avma;
	z[2]=lpile(l,tetpil,gmod(p1,(GEN)z[1]));
	return z;
    }
  }

  /* ici ! is_scalar_t(ty) */
  if ( (vx>vy && (!is_matvec_t(tx) || !is_matvec_t(ty)))
    || (vx==vy && is_scalar_t(tx)) )
  {
    switch(ty)
    {
      case t_POL:
	l=lgef(y);
	if (l==2)
	{
	  if (isexactzero(x))
	  {
	    z=cgetg(2,ty); z[1]=evallgef(2)+evalvarn(vy);
	    return z;
	  }
	  z=cgetg(3,ty); z[2]=lcopy(x);
	  z[1]=(gcmp0(x)) ? evallgef(3) | evalvarn(vy)
	                  : evalsigne(1) | evallgef(3) | evalvarn(vy);
	  return z;
	}

	z=cgetg(l,ty);
	z[2]=ladd(x,(GEN)y[2]);
	if (l==3 && gcmp0((GEN)z[2]))
	{
	  z[1]=isexactzero((GEN)z[2]) ? evallgef(2) | evalvarn(vy)
	                              : evallgef(3) | evalvarn(vy);
	  return z;
	}
	z[1]=y[1]; for (i=3; i<l; i++) z[i]=lcopy((GEN)y[i]);
	normalizepol(&z); return z;
      
      case t_SER:
	l=valp(y);
	if (l<3-ly) return gcopy(y);

	if (l<0)
	{
	  z=cgetg(ly,ty);
	  z[2-l]=ladd(x,(GEN)y[2-l]);
	  z[1]=y[1];
	  for (i=2; i<=1-l; i++) z[i]=lcopy((GEN)y[i]);
	  for (i=3-l; i<ly; i++) z[i]=lcopy((GEN)y[i]);
	  return z;
	}

	if (l>0)
	{
	  if (gcmp0(x)) return gcopy(y);

	  if (gcmp0(y)) lz=l+2; else lz=ly+l;
	  z=cgetg(lz,ty);
	  z[1]=evalsigne(1)+HIGHVALPBIT+evalvarn(vy);
	  z[2]=lcopy(x);
	  for (i=3; i<=l+1; i++) z[i]=zero;
	  for (i=l+2; i<lz; i++) z[i]=lcopy((GEN)y[i-l]);
	  return z;
	}

	z=cgetg(ly,ty); av=avma;
	p1=(signe(y))?gadd(x,(GEN)y[2]):x;
	if (isexactzero(p1))
	{
	  avma=av;
	  if (ly==3)
	  { z[2]=zero; z[1]=HIGHVALPBIT+1+evalvarn(vy); }
	  else
	  {
	    i=3;
	    while (i<ly && gcmp0((GEN)y[i])) i++;
	    if (i==ly)
	    {
	      cgiv(z); z=cgetg(3,ty); z[2]=zero;
	      z[1]=HIGHVALPBIT-2+i+evalvarn(vy);
	    }
	    else
	    {
	      cgiv(z); z=cgetg(ly-i+2,ty);
	      z[1]=evalvalp(i-2)+evalsigne(1)+evalvarn(vy);
	      for (j=2; j<=ly-i+1; j++)
		z[j]=lcopy((GEN)y[j+i-2]);
	    }
	  }
	  return z;
	}

	if (!signe(y)) z[1]=HIGHVALPBIT+evalvarn(vy);
	else
	{
	  z[1]=evalsigne(1)+HIGHVALPBIT+evalvarn(vy);
	  z[2]=(long)p1;
	  for (i=3; i<ly; i++) z[i]=lcopy((GEN)y[i]);
	}
	return z;
      
      case t_RFRAC:
	l=avma; z=cgetg(ly,ty);
	z[1]=ladd(gmul(x,(GEN)y[2]),(GEN)y[1]); z[2]=y[2];
	tetpil=avma; return gerepile(l,tetpil,gred(z));

      case t_RFRACN:
	z=cgetg(ly,ty); l=avma;
	p1=gmul((GEN)y[2],x);
	tetpil=avma;
	z[1]=lpile(l,tetpil,gadd(p1,(GEN)y[1]));
	z[2]=lcopy((GEN)y[2]);
	return z;
      
      case t_VEC: case t_COL:
	if (isexactzero(x)) return gcopy(y); 
	err(gadder5);

      case t_MAT:
	if (isexactzero(x)) return gcopy(y);
	if (ly>=2 && lg(y[1])==ly)
	{
	  l=avma; p1=gscalmat(x,ly-1); tetpil=avma;
	  return gerepile(l,tetpil,gadd(p1,y));
	}
	err(gadder5);

      case t_QFR: case t_QFI:
	err(gadder5);
    }
    err(typeer,"addition");
  }

  /* ici !is_scalar_t(ty) && !is_scalar_t(ty) et vx=vy */
  if (tx>ty) { p1=x; x=y; y=p1; i=tx; tx=ty; ty=i; }
  switch(tx)
  {
    case t_POL:
      switch (ty)
      {
	case t_POL:
	  lx=lgef(x); ly=lgef(y);
	  if (ly>lx) { p1=x; x=y; y=p1; k=lx; lx=ly; ly=k; }
	  z=cgetg(lx,ty);
	  for (i=2; i<ly; i++) z[i]=ladd((GEN)x[i],(GEN)y[i]);
	  for (i=ly; i<lx; i++) z[i]=lcopy((GEN)x[i]);
	  z[1]=x[1]; normalizepol(&z); return z;

	case t_SER:
	  if (gcmp0(x)) return gcopy(y);
	  i=signe(y) ? valp(y)+ly-gval(x,vx) : valp(y)+3-gval(x,vx);
	  if (i<3) return gcopy(y);

	  l=avma; p1=greffe(x,i); tetpil=avma;
	  return gerepile(l,tetpil,gadd(p1,y));
	      
	case t_RFRAC:
	  l=avma; z=cgetg(ly,ty);
	  z[1]=ladd(gmul(x,(GEN)y[2]),(GEN)y[1]); z[2]=y[2];
	  tetpil=avma; return gerepile(l,tetpil,gred(z));

	case t_RFRACN:
	  z=cgetg(ly,ty); l=avma; p1=gmul(x,(GEN)y[2]); tetpil=avma;
	  z[1]=lpile(l,tetpil,gadd(p1,(GEN)y[1])); z[2]=lcopy((GEN)y[2]);
	  return z;

	case t_QFR: case t_QFI:
	case t_VEC: case t_COL: case t_MAT:
	  err(gadder6);
	default: err(typeer,"addition");
      }
    
    case t_SER:
      switch(ty)
      {
	case t_SER:
	  l=valp(y)-valp(x);
	  if (l<0)
	  {
	    l= -l; p1=x; x=y; y=p1; i=lx; lx=ly; ly=i;
	  }
	  if (gcmp0(x)) return gcopy(x);

	  if (gcmp0(y)) ly=2;
	  lz=l+ly; if (lx<lz) lz=lx;
	  if (l)
	  {
	    z=cgetg(lz,ty);
	    z[1]=evalvalp(valp(x))+evalsigne(1)+evalvarn(vx);
	    if (l<lz-2)
	    {
	      for (i=2; i<=l+1; i++) z[i]=lcopy((GEN)x[i]);
	      for (i=l+2; i<lz; i++) z[i]=ladd((GEN)x[i],(GEN)y[i-l]);
	    }
	    else
	      for (i=2; i<lz; i++) z[i]=lcopy((GEN)x[i]);
	    return z;
	  }

	  i=2; l=avma;
	  while (i<lz)
	  {
	    avma=l; p1=gadd((GEN)x[i],(GEN)y[i]);
	    i++;
	    if (!isexactzero(p1))
	    {
	      z=cgetg(lz-i+3,ty);
	      z[1]=evalvalp(valp(x)+i-3)+evalsigne(1)+evalvarn(vx);
	      z[2]=(long)p1;
	      for (j=i; j<lz; j++)
		z[j-i+3]=ladd((GEN)x[j],(GEN)y[j]);
	      return z;
	    }
	  }
	  avma=l; z=cgetg(lz,ty);
	  z[1]=evalvalp(lz-2+valp(y))+evalvarn(vx);
	  return z;
	      
	case t_RFRAC: case t_RFRACN:
	  if (gcmp0(y)) return gcopy(x);

	  l=gval(y,vy);
	  if (gcmp0(x)) lz=valp(x)+3-l;
	  else lz=lx+valp(x)-l;
	  if (lz<3) return gcopy(x);

	  l=avma; ty=typ(y[2]);
	  if (is_scalar_t(ty))
	    p1=gdiv((GEN)y[1], (GEN)y[2]);
	  else
	    p1=gdiv((GEN)y[1], greffe((GEN)y[2],lz));
	  tetpil=avma; return gerepile(l,tetpil,gadd(p1,x));

	case t_QFR: case t_QFI: case t_VEC: case t_COL: case t_MAT:
	  err(gadder7);
	      
	default: err(typeer,"addition");
	      
      } break;
    
    case t_RFRAC: case t_RFRACN:
      if (!is_rfrac_t(ty)) err(gadder10);

      if (tx==t_RFRAC && ty==t_RFRAC) i=t_RFRAC; else i=t_RFRACN;
      av=avma; z=cgetg(ly,i); l=avma;
      p1=gmul((GEN)x[1],(GEN)y[2]);
      p2=gmul((GEN)x[2],(GEN)y[1]);
      tetpil=avma; z[1]=lpile(l,tetpil,gadd(p1,p2));
      z[2]=lmul((GEN)x[2],(GEN)y[2]);
      if (i==t_RFRAC)
      {
	tetpil=avma; p1=gred(z); z=gerepile(av,tetpil,p1);
      }
      return z;
    
    case t_VEC: case t_COL: case t_MAT:
      if (lx!=ly || tx!=ty) err(gadder11);

      z=cgetg(ly,ty);
      for (i=1; i<ly; i++)
	z[i]=ladd((GEN)x[i],(GEN)y[i]);
      return z;
    
    case t_QFR: case t_QFI:
      err(gadder9);
  }
  err(typeer,"addition");
  return NULL; /* not reached */
}

/********************************************************************/
/**                                                                **/
/**                   SOUSTRACTION GENERALE                        **/
/**                                                                **/
/********************************************************************/

GEN
gsub(GEN x, GEN y)
{
  long tetpil, av = avma;
  GEN p1 = gneg(y);

  tetpil=avma; return gerepile(av,tetpil,gadd(x,p1));
}

/********************************************************************/
/**                                                                **/
/**                 MULTIPLICATION GENERALE                        **/
/**                                                                **/
/********************************************************************/

GEN
gmul(GEN x, GEN y)
{
  long tx = typ(x), ty = typ(y), lx,ly,vx,vy,i,j,k,l;
  long av,tetpil;
  GEN z,p1,p2,p3;

  if ( is_const_t(tx) && is_const_t(ty) )
  {
    if (tx>ty) { p1=x; x=y; y=p1; i=tx; tx=ty; ty=i; }
  }
  else
  {
    vx=gvar(x); vy=gvar(y);
    if (!is_matvec_t(ty))
    {
      if (is_matvec_t(tx) || vx<vy || (vx==vy && tx>ty))
      {
	p1=x; x=y; y=p1; i=tx; tx=ty; ty=i;
	i=vx; vx=vy; vy=i;
      }
    }
  }
  lx=lg(x); ly=lg(y);

  if (is_scalar_t(ty)) /* here tx<=ty, or ty=9, tx<17, vx>vy */
  {
    switch(tx)
    {
      case t_INT:
        switch(ty)
	{
	  case t_INT:
	    return mulii(x,y);

	  case t_REAL:
	    return mulir(x,y);

	  case t_INTMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    l=avma; p1=mulii(x,(GEN)y[2]);
	    tetpil=avma; p2=modii(p1,(GEN)y[1]);
	    z[2]=lpile(l,tetpil,p2);
	    return z;

	  case t_FRAC: case t_FRACN:
	    z=cgetg(ly,ty);
	    z[1]=lmulii(x,(GEN)y[1]);
	    z[2]=lcopy((GEN)y[2]);
	    if (ty==t_FRAC) gredsp(&z);
	    return z;
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=lmul(x,(GEN)y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  case t_PADIC:
	    if (signe(x)==0) return gzero;
	    l=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
	    return gerepile(l,tetpil,gmul(p1,y));
	  
	  case t_QUAD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    z[3]=lmul(x,(GEN)y[3]);
	    return z;
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[2]=lmul(x,(GEN)y[2]);
	    z[1]=copyifstack(y[1]);
	    return z;
	  
	  default: err(typeer,"multiplication");
	}
	
      case t_REAL:
        switch(ty)
	{
	  case t_REAL:
	    return mulrr(x,y);

	  case t_FRAC: case t_FRACN:
	    l=avma; p1=cgetr(lx);
	    tetpil=avma; gaffect(y,p1);
	    p2=mulrr(p1,x); return gerepile(l,tetpil,p2);
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=lmul(x,(GEN)y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  case t_QUAD:
	    l=avma; p1=co8(y,lx); tetpil=avma;
	    return gerepile(l,tetpil,gmul(p1,x));
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  case t_INTMOD:
	    err(gmuler2);
	  default: err(typeer,"multiplication");
	}
	
      case t_INTMOD:
        switch(ty)
	{
	  case t_INTMOD:
	    z=cgetg(ly,ty); p2=(GEN)x[1]; p3=(GEN)y[1];
	    if (p2==p3 || gegal(p2,p3)) z[1]=copyifstack(p2);
	    else z[1]=lmppgcd(p2,p3);
	    l=avma; p1=mulii((GEN)x[2],(GEN)y[2]); tetpil=avma;
	    z[2]=lpile(l,tetpil,modii(p1,(GEN)z[1]));
	    return z;
	  
	  case t_FRAC: case t_FRACN:
	    z=cgetg(3,t_INTMOD);
	    z[1]=copyifstack(x[1]);
	    z[2]=lgeti(lgefint(x[1]));
	    gaffect(y,z); gmulz(z,x,z);
	    return z;
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=lmul(x,(GEN)y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  case t_PADIC:
	    l=avma; p1=cgetg(3,t_INTMOD);
	    p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
	    gaffect(y,p1); tetpil=avma;
	    return gerepile(l,tetpil,gmul(x,p1));

	  case t_QUAD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]); z[3]=lmul(x,(GEN)y[3]);
	    return z;
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  default: err(typeer,"multiplication");
	}
	
      case t_FRAC: case t_FRACN:
        switch(ty)
	{
	  case t_FRAC: case t_FRACN:
	    z=cgetg(ly,t_FRACN);
	    z[1]=lmulii((GEN)x[1],(GEN)y[1]);
	    z[2]=lmulii((GEN)x[2],(GEN)y[2]);
	    if (tx==t_FRAC && ty==t_FRAC) gredsp(&z);
	    return z;
	  
	  case t_COMPLEX:
	    z=cgetg(ly,ty);
	    z[1]=lmul((GEN)y[1],x);
	    z[2]=lmul((GEN)y[2],x);
	    return z;
	  
	  case t_PADIC:
	    if (signe(x[1]) == 0) return gzero;

	    l=avma; p1=cgetp(y); gaffect(x,p1);
	    tetpil=avma; return gerepile(l,tetpil,gmul(p1,y));
	  
	  case t_QUAD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul((GEN)y[2],x);
	    z[3]=lmul((GEN)y[3],x);
	    return z;
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  default: err(typeer,"multiplication");
	}
	
      case t_COMPLEX:
        switch(ty)
	{
	  case t_COMPLEX:
	    z=cgetg(ly,ty); l=avma;
	    p1=gmul((GEN)x[1],(GEN)y[1]); p2=gmul((GEN)x[2],(GEN)y[2]);
	    x=gadd((GEN)x[1],(GEN)x[2]); y=gadd((GEN)y[1],(GEN)y[2]);
	    y=gmul(x,y); x=gadd(p1,p2);
	    tetpil=avma; z[1]=lsub(p1,p2); z[2]=lsub(y,x);
	    gerepilemanyvec(l,tetpil,z+1,2);
	    return z;
	  
	  case t_PADIC:
	    p1=(GEN)y[2];
	    if (mod4(p1) == 3) /* we suppose p1>0 */
	    {
	      z=cgetg(3,t_COMPLEX);
	      z[1]=lmul((GEN)x[1],y);
	      z[2]=lmul((GEN)x[2],y);
	      return z;
	    }

	    l=avma;
	    if (signe(y[4])) p1=cvtop(x,(GEN)y[2],precp(y));
	    else p1=cvtop(x,(GEN)y[2],(valp(y)>0)?valp(y)+1:1);
	    tetpil=avma; return gerepile(l,tetpil,gmul(p1,y));
	  
	  case t_QUAD:
	    lx=precision(x); if (!lx) err(gmuler11);
	    l=avma; p1=co8(y,lx); tetpil=avma;
	    return gerepile(l,tetpil,gmul(p1,x));
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  default: err(gmuler11);
	}
	
      case t_PADIC:
        switch(ty)
	{
	  case t_PADIC:
	    if (cmpii((GEN)x[2],(GEN)y[2])) err(gmuler14);
	    if (!signe(x[4]))
	      { z=gcopy(x); setvalp(z,valp(x)+valp(y)); return z; }
	    if (!signe(y[4]))
	      { z=gcopy(y); setvalp(z,valp(x)+valp(y)); return z; }

	    p1 = (precp(x)>precp(y)) ? y : x;
	    z=cgetp(p1); l=avma; setvalp(z,valp(x)+valp(y));
	    modiiz(mulii((GEN)x[4],(GEN)y[4]),(GEN)p1[3],(GEN)z[4]);
	    avma=l; return z;
	  
	  case t_QUAD:
	    if (kro8(y,(GEN)x[2])== -1)
	    {
	      z=cgetg(4,t_QUAD);
	      z[1]=copyifstack(y[1]);
	      z[2]=lmul((GEN)y[2],x);
	      z[3]=lmul((GEN)y[3],x);
	      return z;
	    }

	    l=avma; p1=cvtop(y,(GEN)x[2],signe(x[4])?precp(x):valp(x)+1);
	    tetpil=avma; return gerepile(l,tetpil,gmul(p1,x));
	  
	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  default: err(typeer,"multiplication");
	}
	
      case t_QUAD:
        switch(ty)
	{
	  case t_QUAD:
	  {
	    GEN p4,yy=(GEN)y[1];
	  
	    p1=(GEN)x[1]; 
	    if (p1!=yy && !gegal(p1,yy)) err(gmuler12);
	    z=cgetg(ly,ty); l=avma;
	    p2=gmul((GEN)x[2],(GEN)y[2]);
	    p3=gmul((GEN)x[3],(GEN)y[3]);
	    p4=gmul(gneg((GEN)p1[2]),p3); tetpil=avma;

	    if (gcmp0((GEN)p1[3]))
	    {
	      z[2]=lpile(l,tetpil,gadd(p4,p2));
	      l=avma;
	      p2=gmul((GEN)x[2],(GEN)y[3]);
	      p3=gmul((GEN)x[3],(GEN)y[2]);
	      tetpil=avma;
	      z[3]=lpile(l,tetpil,gadd(p2,p3));
	      z[1]=copyifstack(yy); return z;
	    }

	    p1=gadd( gmul((GEN)x[2],(GEN)y[3]), gmul((GEN)x[3],(GEN)y[2]) );
	    tetpil=avma;
	    z[2]=ladd(p2,p4); z[3]=ladd(p1,p3);
	    gerepilemanyvec(l,tetpil,z+2,2);
	    z[1]=copyifstack(yy); return z;
	  }

	  case t_POLMOD:
	    z=cgetg(ly,ty);
	    z[1]=copyifstack(y[1]);
	    z[2]=lmul(x,(GEN)y[2]);
	    return z;
	  
	  default: err(typeer,"multiplication");
	}	
	
      case t_POLMOD:
        z=cgetg(lx,tx); k=x[1]; l=y[1];
	if (k==l || gegal((GEN)k,(GEN)l))
	{ z[1]=copyifstack(k); l=avma; p1=gmul((GEN)x[2],(GEN)y[2]); }
	else
	{
	  vx=varn(x[1]); vy=varn(y[1]);
	  if (vx==vy)
	  { z[1]=lgcd((GEN)k,(GEN)l); l=avma; p1=gmul((GEN)x[2],(GEN)y[2]); }
	  else
	  {
	    if (vx<vy)
	    { z[1]=copyifstack(k); l=avma; p1=gmul((GEN)x[2],y); }
	    else
	    { z[1]=copyifstack(l); l=avma; p1=gmul((GEN)y[2],x); }
	  }
	}
	tetpil=avma; z[2]=lpile(l,tetpil,gmod(p1,(GEN)z[1]));
	return z;
	
      default: if (ty!=t_POLMOD) err(gmulbug1);
	z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	l=avma; p1=gmul(x,(GEN)y[2]); tetpil=avma;
	z[2]=lpile(l,tetpil,gmod(p1,(GEN)z[1]));
	return z;
    }
  }

  /* ici ty>=10 */
  if ( ( vx>vy && (!is_matvec_t(tx) || !is_matvec_t(ty)) )
    || (vx==vy && is_scalar_t(tx)) || (is_matvec_t(ty) && !is_matvec_t(tx)) )
  {
    switch(ty)
    {
      case t_POL:
        z=cgetg(ly,ty);
	if (isexactzero(x) || isexactzero(y))
	{ z[1]=evallgef(2)+evalvarn(vy); }
	else
	{
	  for (i=2; i<lgef(y); i++) z[i]=lmul(x,(GEN)y[i]);
	  z[1]=y[1]; normalizepol(&z);
	}
	return z;
	
      case t_SER:
	if (isexactzero(x))
	{ 
	  z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vy);
	  return z;
	}

	if (!signe(y)) return gcopy(y);

	z=cgetg(ly,ty);
	for (i=2; i<ly; i++) z[i]=lmul(x,(GEN)y[i]);
	z[1]=y[1]; normalize(&z); return z;
	
      case t_RFRAC:
	if (isexactzero(x))
	{ 
	  z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vy);
	  return z;
	}

	l=avma; z=cgetg(ly,ty); z[1]=lmul(x,(GEN)y[1]); z[2]=y[2];
	tetpil=avma; return gerepile(l,tetpil,gred(z));

      case t_RFRACN:
	if (isexactzero(x)) 
	{ 
	  z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vy);
	  return z;
	}
	z=cgetg(ly,ty); z[1]=lmul(x,(GEN)y[1]); z[2]=lcopy((GEN)y[2]);
	return z;
	
      case t_VEC: case t_COL: case t_MAT:
        z=cgetg(ly,ty);
	for (i=1; i<ly; i++) z[i]=lmul(x,(GEN)y[i]);
	return z;
    }
    err(typeer,"multiplication");
  }

  /* ici ty>=10 et tx>=10 et vx=vy ou tx et ty>=17 */
  if ( tx>ty && (!is_matvec_t(tx) || !is_matvec_t(ty)) )
  {
    p1=x; x=y; y=p1; i=tx; tx=ty; ty=i;
    i=lx; lx=ly; ly=i;
  }

  switch(tx)
  {
    case t_POL:
      switch (ty)
      {
	case t_POL:
	  if (isexactzero(x) || isexactzero(y))
	  {
	    z=cgetg(2,ty); z[1]=evallgef(2)+evalvarn(vx);
	    return z;
	  }

	  lx=lgef(x); ly=lgef(y);
	  if (lx<ly)
	  {
	    p1=x; x=y; y=p1; k=lx; lx=ly; ly=k;
	  }
	  k=lx+ly-3; z=cgetg(k,ty);
	  for (i=2; i<ly; i++)
	  {
	    p1=gzero; l=avma;
	    for (j=2; j<=i; j++)
	    {
	      p2=gmul((GEN)y[j],(GEN)x[i-j+2]);
	      tetpil=avma;
	      p1=gadd(p1,p2);
	    }
	    z[i]=lpile(l,tetpil,p1);
	  }
	  for ( ; i<lx; i++)
	  {
	    p1=gzero; l=avma;
	    for (j=2; j<ly; j++)
	    {
	      p2=gmul((GEN)y[j],(GEN)x[i-j+2]);
	      tetpil=avma;
	      p1=gadd(p1,p2);
	    }
	    z[i]=lpile(l,tetpil,p1);
	  }
	  for ( ; i<=lx+ly-4; i++)
	  {
	    p1=gzero; l=avma;
	    for (j=i-lx+3; j<ly; j++)
	    {
	      p2=gmul((GEN)y[j],(GEN)x[i-j+2]);
	      tetpil=avma;
	      p1=gadd(p1,p2);
	    }
	    z[i]=lpile(l,tetpil,p1);
	  }
	  z[1]=evalsigne(1)+evallgef(k)+evalvarn(vx);
	  normalizepol(&z); return z;

	case t_SER:
	  if (gcmp0(x))
	  {
	    z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vx);
	    return z;
	  }

	  if (gcmp0(y))
	  {
	    z=cgetg(3,t_SER);
	    z[1]=evalvalp(valp(y)+gval(x,vx))+evalvarn(vx);
	    return z;
	  }
	  l=avma; p1=greffe(x,ly); tetpil=avma; p2=gmul(p1,y);
	  return gerepile(l,tetpil,p2);
	  
	case t_RFRAC: case t_RFRACN:
	  l=avma; z=cgetg(ly,ty); z[1]=lmul(x,(GEN)y[1]);
	  z[2]=lcopy((GEN)y[2]);
	  if (ty==t_RFRAC)
	  {
	    tetpil=avma; z=gerepile(l,tetpil,gred(z));
	  }
	  return z;
	  
	case t_VEC: case t_COL: case t_MAT:
	  z=cgetg(ly,ty);
	  for (i=1; i<ly; i++) z[i]=lmul(x,(GEN)y[i]);
	  return z;
	  
	default: err(typeer,"multiplication");
      }
	
    case t_SER:
      switch (ty)
      {
	case t_SER:
	  if (lx>ly) { k=ly; ly=lx; lx=k; p1=y; y=x; x=p1; }
	  if (gcmp0(x)||gcmp0(y))
	  {
	    z=cgetg(3,t_SER); z[1]=evalvalp(valp(x)+valp(y))+evalvarn(vx);
	    return z;
	  }

	  l=avma; p1=cgeti(lx); z=cgetg(lx,ty);
	  z[1]=evalvalp(valp(x)+valp(y))+evalvarn(vx)+evalsigne(1);
	  for (i=2; i<lx; i++)
	  {
	    p1[i]=!isexactzero((GEN)y[i]);
	    z[i]=p1[i]?lmul((GEN)x[2],(GEN)y[i]):zero;
	  }
	  for (i=3; i<lx; i++)
	    if (!isexactzero((GEN)x[i]))
	      for (j=2; j<=lx+1-i; j++)
		if (p1[j])
		  z[i+j-2]=ladd((GEN)z[i+j-2],gmul((GEN)x[i],(GEN)y[j]));
	  tetpil=avma; z=gerepile(l,tetpil,gcopy(z));
	  normalize(&z); return z;
	  
	case t_RFRAC: case t_RFRACN:
	  if (gcmp0(y))
	  {
	    z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vx);
	    return z;
	  }

	  if (gcmp0(x))
	  {
	    z=cgetg(3,t_SER);
	    z[1]=evalvalp(valp(x)+gval(y,vx))+evalvarn(vx);
	    return z;
	  }

	  l=avma; p1=gmul((GEN)y[1],x);
	  tetpil=avma; return gerepile(l,tetpil,gdiv(p1,(GEN)y[2]));
	  
	case t_VEC: case t_COL: case t_MAT:
	  z=cgetg(ly,ty);
	  for (i=1; i<ly; i++) z[i]=lmul(x,(GEN)y[i]);
	  return z;
	  
	default: err(typeer,"multiplication");
      }
	
    case t_RFRAC: case t_RFRACN:
      if (is_rfrac_t(ty))
      {
	if (tx == t_RFRAC && ty == t_RFRAC) i=t_RFRAC; else i=t_RFRACN;
	l=avma; z=cgetg(ly,i);
	z[1]=lmul((GEN)x[1],(GEN)y[1]);
	z[2]=lmul((GEN)x[2],(GEN)y[2]);
	if (i==t_RFRAC)
	{
	  tetpil=avma; z=gerepile(l,tetpil,gred(z));
	}
	return z;
      }

      if (!is_matvec_t(ty)) err(typeer,"multiplication");

      z=cgetg(ly,ty);
      for (i=1; i<ly; i++) z[i]=lmul(x,(GEN)y[i]);
      return z;

    case t_QFR:
      if (ty!=t_QFR) err(typeer,"multiplication");
      return compreal(x,y);
	
    case t_QFI:
      if (ty!=t_QFI) err(typeer,"multiplication");
      return compimag(x,y);
	
    case t_VEC:
      switch(ty)
      {
	case t_COL:
	  if (lx!=ly) err(gmuler7);

	  z=gzero; l=avma;
	  for (i=1; i<lx; i++)
	  {
	    p1=gmul((GEN)x[i],(GEN)y[i]);
	    tetpil=avma;
	    z=gadd(z,p1);
	  }
	  return gerepile(l,tetpil,z);
	  
	case t_MAT:
	  if (ly==1) return cgetg(1,t_VEC);

	  l=lg(y[1]); if (lx!=l) err(gmuler8);

	  z=cgetg(ly,tx);
	  for (i=1; i<ly; i++)
	  {
	    p1=gzero; av=avma;
	    for (j=1; j<lx; j++)
	    {
	      p2=gmul((GEN)x[j],gcoeff(y,j,i));
	      tetpil=avma; p1=gadd(p1,p2);
	    }
	    z[i]=lpile(av,tetpil,p1);
	  }
	  return z;

	case t_VEC:
	  err(gmuler7);
	default: err(typeer,"multiplication");
      }
	
    case t_COL:
      switch(ty)
      {
	case t_VEC:
	  z=cgetg(ly,t_MAT);
	  for (i=1; i<ly; i++) z[i]=lmul((GEN)y[i],x);
	  return z;
	  
	case t_MAT:
	  if (ly!=1 && lg(y[1])!=2) err(gmuler7);

	  z=cgetg(ly,t_MAT);
	  for (i=1; i<ly; i++) z[i]=lmul(gcoeff(y,1,i),x);
	  return z;
	  
	case t_COL:
	  err(gmuler7);
	default: err(typeer,"multiplication");
      }
	
    case t_MAT:
      switch(ty)
      {
	case t_VEC:
	  if (lx!=2) err(gmuler7);
	  z=cgetg(ly,t_MAT);
	  for (i=1; i<ly; i++) z[i]=lmul((GEN)y[i],(GEN)x[1]);
	  return z;
	  
	case t_COL:
	  if (lx!=ly) err(gmuler9);
	  if (lx==1) return gcopy(y);

	  lx=lg(x[1]); z=cgetg(lx,ty);
	  for (i=1; i<lx; i++)
	  {
	    p1=gzero; l=avma;
	    for (j=1; j<ly; j++)
	    {
	      p2=gmul(gcoeff(x,i,j),(GEN)y[j]);
	      tetpil=avma; p1=gadd(p1,p2);
	    }
	    z[i]=lpile(l,tetpil,p1);
	  }
	  return z;
	  
	case t_MAT:
	  if (lx==1)
	  {
	    if (ly!=1 && lg(y[1])!=1) err(gmuler10);
	    if (ly>1)
	    {
	      z=cgetg(ly,tx);
	      for (i=1; i<ly; i++) z[i]=lgetg(1,t_COL);
	      return z;
	    }
	  }
	  if (ly==1) return cgetg(ly,tx);

	  l=lg(y[1]); if (lx!=l) err(gmuler10);
	  l=lg(x[1]); z=cgetg(ly,tx);
	  for (i=1; i<ly; i++) z[i]=lgetg(l,t_COL);
	  for (i=1; i<l; i++)
	  {
	    for (j=1; j<ly; j++)
	    {
	      p1=gzero; av=avma;
	      for (k=1; k<lx; k++)
	      {
		p2=gmul(gcoeff(x,i,k),gcoeff(y,k,j));
		tetpil=avma; p1=gadd(p1,p2);
	      }
	      coeff(z,i,j)=lpile(av,tetpil,p1);
	    }
	  }
	  return z;
	  
	default: err(typeer,"multiplication");
      }
  }
  err(typeer,"multiplication");
  return NULL; /* not reached */
}

/********************************************************************/
/**                                                                **/
/**                       DIVISION GENERALE                        **/
/**                                                                **/
/********************************************************************/

GEN
gdiv(GEN x, GEN y)
{
  long lx,ly,tx,ty,vx,vy,i,j,k,l;
  long av,tetpil;
  GEN z,p1,p2,p3;

  if (gcmp0(y)) err(gdiver2);

  lx=lg(x); ly=lg(y); tx=typ(x); ty=typ(y);
  if (is_scalar_t(tx) && is_scalar_t(ty))
  {
    switch(tx)
    {
      case t_INT:
	switch(ty)
	{
	  case t_INT:
	    i=signe(x); if (!i) return gzero;
	    l=avma; z=dvmdii(x,y,&p1);
	    if (!signe(p1)) { cgiv(p1); return z; }

	    p1=mppgcd(y,p1); tetpil=avma;
	    z=cgetg(3,t_FRAC);
	    if (!gcmp1(p1))
	    {
	      z[1]=ldivii(x,p1);
	      z[2]=ldivii(y,p1);
	    } else { z[1]=lcopy(x); z[2]=lcopy(y); }
	    if (signe(y)<0)
	    {
	      setsigne(z[1],-i);
	      setsigne(z[2],1);
	    }
	    return gerepile(l,tetpil,z);
		
	  case t_REAL:
	    return divir(x,y);
		
	  case t_INTMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=mpinvmod((GEN)y[2],(GEN)y[1]);
	    p2=mulii(x,p1); tetpil=avma;
	    z[2]=lpile(l,tetpil,modii(p2,(GEN)y[1]));
	    return z;
		
	  case t_FRAC: case t_FRACN:
	    z=cgetg(ly,ty);
	    z[1]=lmulii(x,(GEN)y[2]);
	    z[2]=lcopy((GEN)y[1]);
	    if (signe(z[2])<0)
	    {
	      setsigne(z[1],-signe(z[1]));
	      setsigne(z[2],1);
	    }
	    if (ty==t_FRAC) gredsp(&z);
	    return z;
		
	  case t_PADIC:
	    if (signe(x))
	    {
	      l=avma; p1=cgetp(y); gaffect(x,p1);
	      tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));
	    }
	    return gzero;
		
	  case t_COMPLEX: case t_QUAD:
	    l=avma; p1=gnorm(y);
	    p2=gmul(x,gconj(y)); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p2,p1));
		
	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma; p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;
		
	  default: err(typeer,"division");
	}
      
      case t_REAL:
	switch(ty)
	{
	  case t_INT:
	    return divri(x,y);

	  case t_REAL:
	    return divrr(x,y);

	  case t_FRAC: case t_FRACN:
	    l=avma; p1=cgetg(lx,tx); gaffect(y,p1);
	    p2=divrr(x,p1); return gerepile(l,(long)p1,p2);

	  case t_COMPLEX:
	    z=cgetg(ly,ty); l=avma; p1=gnorm(y);
	    p2=gmul(x,(GEN)y[1]); p3=gmul(x,(GEN)y[2]);
	    if (!gcmp0(p3)) gnegz(p3,p3); tetpil=avma;
	    z[1]=ldiv(p2,p1); z[2]=ldiv(p3,p1);
	    gerepilemanyvec(l,tetpil,z+1,2); return z;

	  case t_QUAD:
	    l=avma; p1=co8(y,lx); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(x,p1));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  case t_INTMOD: case t_PADIC:
	    err(gdiver3);
	  default: err(typeer,"division");
	}

      case t_INTMOD:
	switch(ty)
	{
	  case t_INT:
	    z=cgetg(lx,tx); z[1]=copyifstack(x[1]);
	    z[2]=lmpinvmod(y,(GEN)x[1]); l=avma;
	    p1=mulii((GEN)x[2],(GEN)z[2]);
	    mpmodz(p1,(GEN)x[1],(GEN)z[2]); avma=l;
	    return z;

	  case t_INTMOD:
	    z=cgetg(ly,ty); k=x[1]; l=y[1];
	    if (k==l || gegal((GEN)k,(GEN)l)) z[1]=copyifstack(k);
	    else z[1]=lmppgcd((GEN)k,(GEN)l);
	    l=avma; p1=mpinvmod((GEN)y[2],(GEN)z[1]);
	    p2=mulii((GEN)x[2],p1); tetpil=avma;
	    z[2]=lpile(l,tetpil,modii(p2,(GEN)z[1]));
	    return z;

	  case t_FRAC:
	    z=cgetg(lx,tx);
	    z[2]=lmpinvmod((GEN)y[1],(GEN)x[1]);
	    l=avma;
	    p1=mulii((GEN)x[2],(GEN)y[2]); mpmodz(p1,(GEN)x[1],p1);
	    p2=mulii(p1,(GEN)z[2]); mpmodz(p2,(GEN)x[1],(GEN)z[2]);
	    avma=l; z[1]=copyifstack(x[1]);
	    return z;

	  case t_FRACN:
	    l=avma; p1=gred(y); tetpil=avma;
	    p2=gdiv(x,p1); return gerepile(l,tetpil,p2);

	  case t_COMPLEX: case t_QUAD:
	    l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p2,p1));

	  case t_PADIC:
	    l=avma; p1=cgetg(3,t_INTMOD); p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
	    gaffect(y,p1); tetpil=avma; return gerepile(l,tetpil,gdiv(x,p1));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma; p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  case t_REAL:
	    err(gdiver4);
	  default: err(typeer,"division");
	}

      case t_FRAC: case t_FRACN:
	switch(ty)
	{
	  case t_INT:
	    z=cgetg(lx,tx);
	    z[1]=lcopy((GEN)x[1]);
	    z[2]=lmul((GEN)x[2],y);
	    if (signe(z[2])<0)
	    {
	      setsigne(z[1],-signe(z[1]));
	      setsigne(z[2],1);
	    }
	    if (tx==t_FRAC) gredsp(&z);
	    return z;

	  case t_REAL:
	    l=avma; p1=cgetg(ly,ty); gaffect(x,p1);
	    p2=divrr(p1,y); return gerepile(l,(long)p1,p2);

	  case t_INTMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=mulii((GEN)y[2],(GEN)x[2]);
	    p1=mpinvmod(p1,(GEN)y[1]);
	    p1=mulii(p1,(GEN)x[1]); tetpil=avma;
	    z[2]=lpile(l,tetpil,modii(p1,(GEN)y[1]));
	    return z;

	  case t_FRAC: case t_FRACN:
	    if (tx == t_FRAC && ty == t_FRAC) i=t_FRAC; else i=t_FRACN;
	    z=cgetg(ly,i);
	    z[1]=lmulii((GEN)x[1],(GEN)y[2]);
	    z[2]=lmulii((GEN)x[2],(GEN)y[1]);
	    if (signe(z[2])<0)
	    {
	      setsigne(z[1],-signe(z[1]));
	      setsigne(z[2],1);
	    }
	    if (i==t_FRAC) gredsp(&z);
	    return z;

	  case t_COMPLEX:
	    z=cgetg(ly,ty); l=avma; p1=gnorm(y);
	    p2=gmul(x,(GEN)y[1]);
	    p3=gmul(x,(GEN)y[2]); if(!gcmp0(p3)) gnegz(p3,p3);
	    tetpil=avma;
	    z[1]=ldiv(p2,p1); z[2]=ldiv(p3,p1);
	    gerepilemanyvec(l,tetpil,z+1,2);
	    return z;

	  case t_PADIC:
	    if (signe(x[1]) == 0) return gzero;

	    l=avma; p1=cgetp(y); gaffect(x,p1);
	    tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));

	  case t_QUAD:
	    l=avma; p1=gnorm(y);
	    p2=gmul(x,gconj(y)); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p2,p1));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  default: err(typeer,"division");
	}

      case t_COMPLEX:
	switch(ty)
	{
	  case t_INT: case t_REAL: case t_INTMOD: case t_FRAC: case t_FRACN:
	    z=cgetg(lx,tx);
	    z[1]=ldiv((GEN)x[1],y);
	    z[2]=ldiv((GEN)x[2],y);
	    return z;

	  case t_COMPLEX:
	    l=avma; p1=gnorm(y); p2=gconj(y);
	    p2=gmul(x,p2); tetpil=avma;
	    p1=gdiv(p2,p1);
	    return gerepile(l,tetpil,p1);

	  case t_PADIC:
	    if (krosg(-1,(GEN)y[2])== -1)
	    {
	      z=cgetg(3,t_COMPLEX);
	      z[1]=ldiv((GEN)x[1],y);
	      z[2]=ldiv((GEN)x[2],y);
	      return z;
	    }
	    l=avma; p1=cvtop(x,(GEN)y[2],precp(y));
	    tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p1,y));

	  case t_QUAD:
	    lx=precision(x); if (!lx) err(gdiver7);
	    l=avma; p1=co8(y,lx); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(x,p1));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  default: err(typeer,"division");
	}

      case t_PADIC:
	switch(ty)
	{
	  case t_INT: case t_FRAC: case t_FRACN:
	    l=avma;
	    if (signe(x[4])) { p1=cgetp(x); gaffect(y,p1); }
	    else p1=cvtop(y,(GEN)x[2],(valp(x)>0)?valp(x):1);
	    tetpil=avma; return gerepile(l,tetpil,gdiv(x,p1));

	  case t_INTMOD:
	    l=avma; p1=cgetg(3,t_INTMOD);
	    p1[1]=y[1]; p1[2]=lgeti(lg(y[1]));
	    gaffect(x,p1); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p1,y));

	  case t_PADIC:
	    if (cmpii((GEN)x[2],(GEN)y[2])) err(gdiver19);
	    if (!signe(x[4]))
	    { 
	      z=gcopy(x); setvalp(z,valp(x)-valp(y));
	      return z;
	    }

	    p1=(precp(x)>precp(y)) ? y : x;
	    z=cgetp(p1); l=avma;
	    setvalp(z,valp(x)-valp(y));
	    p2=mpinvmod((GEN)y[4],(GEN)p1[3]);
	    modiiz(mulii((GEN)x[4],p2),(GEN)p1[3],(GEN)z[4]);
	    avma=l; return z;

	  case t_COMPLEX: case t_QUAD:
	    l=avma; p1=gmul(x,gconj(y));
	    p2=gnorm(y); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p1,p2));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  case t_REAL:
	    err(talker,"forbidden division p-adic/R");
	  default: err(typeer,"division");
	}

      case t_QUAD:
	switch (ty)
	{
	  case t_INT: case t_INTMOD: case t_FRAC: case t_FRACN:
	    z=cgetg(lx,tx);
	    z[1]=copyifstack(x[1]);
	    for (i=2; i<4; i++) z[i]=ldiv((GEN)x[i],y);
	    return z;

	  case t_REAL:
	    l=avma; p1=co8(x,ly); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p1,y));

	  case t_PADIC:
	    l=avma; p1=cvtop(x,(GEN)y[2],precp(y));
	    tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));

	  case t_COMPLEX:
	    ly=precision(y); if (!ly) err(gdiver6);
	    l=avma; p1=co8(x,ly); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p1,y));

	  case t_QUAD:
	    k=x[1]; l=y[1];
	    if (k!=l && !gegal((GEN)k,(GEN)l)) err(gdiver18);
	    l=avma; p1=gnorm(y);
	    p2=gmul(x,gconj(y)); tetpil=avma;
	    return gerepile(l,tetpil,gdiv(p2,p1));

	  case t_POLMOD:
	    z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	    l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)y[1]);
	    tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	    return z;

	  default: err(typeer,"division");
	}
      case t_POLMOD:
	if (ty!=t_POLMOD)
	{
	  z=cgetg(lx,tx);
	  z[1]=copyifstack(x[1]);
	  z[2]=ldiv((GEN)x[2],y);
	  return z;
	}

	z=cgetg(ly,ty); k=x[1]; l=y[1];
	if (k==l || gegal((GEN)k,(GEN)l))
	{
	  z[1]=copyifstack(k); l=avma;
	  p1=ginvmod((GEN)y[2],(GEN)z[1]); p2=gmul((GEN)x[2],p1);
	}
	else
	{
	  vx=varn(x[1]); vy=varn(y[1]);
	  if (vx==vy)
	  {
	    z[1]=lgcd((GEN)k,(GEN)l); l=avma;
	    p1=ginvmod((GEN)y[2],(GEN)z[1]);
	    p2=gmul((GEN)x[2],p1);
	  }
	  else
	  {
	    if (vx<vy)
	    { z[1]=copyifstack(k); l=avma; p2=gdiv((GEN)x[2],y); }
	    else
	    { z[1]=copyifstack(l); l=avma; p2=gdiv((GEN)y[2],x); }
	  }
	}
	tetpil=avma; z[2]=lpile(l,tetpil,gmod(p2,(GEN)z[1]));
	return z;
    }
    err(typeer,"division");
  }

  vx=gvar(x); vy=gvar(y);
  if (  (vx<vy && (!is_matvec_t(tx) || !is_matvec_t(ty)))
     || (vx==vy && is_scalar_t(ty)) || (is_matvec_t(tx) && !is_matvec_t(ty)))
  {
    l=avma; z=cgetg(lx,tx);
    switch(tx)
    {
      case t_POLMOD:
	z[1]=copyifstack(x[1]); z[2]=ldiv((GEN)x[2],y); return z;

      case t_POL:
	l=lgef(x);
	for (i=2; i<l; i++)
	  z[i]=ldiv((GEN)x[i],y);
	z[1]=x[1]; return z;

      case t_SER:
	if (gcmp0(x)) return gcopy(x);

	for (i=2; i<lx; i++)
	  z[i]=ldiv((GEN)x[i],y);
	setvalp(z,valp(x)); setvarn(z,vx);
	normalize(&z); return z;

      case t_RFRAC:
	z[1]=x[1]; z[2]=lmul((GEN)x[2],y);
	tetpil=avma; return gerepile(l,tetpil,gred(z));

      case t_RFRACN:
	z[1]=lcopy((GEN)x[1]); z[2]=lmul((GEN)x[2],y); return z;

      case t_VEC: case t_COL: case t_MAT:
	for (i=1; i<lx; i++) z[i]=ldiv((GEN)x[i],y);
	return z;
    }
    err(typeer,"division");
  }

  if (vy<vx || (vy==vx && is_scalar_t(tx)))
  {
    switch(ty)
    {
      case t_POLMOD:
	z=cgetg(ly,ty); z[1]=copyifstack(y[1]);
	l=avma; p1=ginvmod((GEN)y[2],(GEN)y[1]);
	tetpil=avma; z[2]=lpile(l,tetpil,gmul(x,p1));
	return z;

      case t_POL:
	if (lgef(y)>3)
	{
	  if (isexactzero(x))
	  { 
	    z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vy);
	    return z;
	  }

	  l=avma; z=cgetg(3,t_RFRAC); z[1]=(long)x;
	  z[2]=(long)y; tetpil=avma;
	  return gerepile(l,tetpil,gred(z));
	}
	return gdiv(x,(GEN)y[2]);

      case t_SER:
	l=avma;
	if (gcmp0(x))
	{
	  p1=ginv(y); tetpil=avma; /* a ameliorer !!!! */
	  return gerepile(l,tetpil,gmul(x,p1));
	}

	p1=cgetg(ly,ty); p1[2]=lcopy(x);
	p1[1]=evalsigne(1)+HIGHVALPBIT+evalvarn(vy);
	for (i=3; i<ly; i++) p1[i]=zero;
	tetpil=avma; p2=gdiv(p1,y);
	return gerepile(l,tetpil,p2);

      case t_RFRAC:
	l=avma; z=cgetg(ly,ty); z[1]=lmul(x,(GEN)y[2]); z[2]=y[1];
	tetpil=avma; return gerepile(l,tetpil,gred(z));

      case t_RFRACN:
	z=cgetg(ly,ty); z[1]=lmul(x,(GEN)y[2]); z[2]=lcopy((GEN)y[1]);
	return z;

      case t_MAT:
	if (ly==1 || lg(y[1])!=ly) err(gdiver9);
	l=avma; p1=invmat(y); tetpil=avma;
	return gerepile(l,tetpil,gmul(x,p1));

      case t_VEC: case t_COL:
	err(gdiver8);
    }
    err(typeer,"division");
  }

  /* ici vx=vy et tx>=10 et ty>=10*/
  switch(tx)
  {
    case t_POL:
      switch(ty)
      {
	case t_POL:
	  l=avma; z=cgetg(3,t_RFRAC);
	  z[1]=(long)x; z[2]=(long)y; tetpil=avma;
	  return gerepile(l,tetpil,gred(z));

	case t_SER:
	  if (gcmp0(x))
	  {
	    z=cgetg(2,t_POL); z[1]=evallgef(2)+evalvarn(vx);
	    return z;
	  }

	  l=avma; p1=greffe(x,ly);
	  tetpil=avma; p2=gdiv(p1,y);
	  return gerepile(l,tetpil,p2);

	case t_RFRAC: case t_RFRACN:
	  l=avma; z=cgetg(ly,ty);
	  z[1]=lmul(x,(GEN)y[2]);
	  if (ty!=t_RFRAC) { z[2]=lcopy((GEN)y[1]); return z;}
	  z[2]=y[1]; tetpil=avma; return gerepile(l,tetpil,gred(z));

	case t_VEC: case t_COL: case t_MAT:
	  err(gdiver10);
	default: err(typeer,"division");
      }

    case t_SER:
      switch(ty)
      {
	case t_POL:
	  l=avma; p1=greffe(y,lx); tetpil=avma;
	  p2=gdiv(x,p1); return gerepile(l,tetpil,p2);

	case t_SER:
	  l=lx; if (ly<lx) l=ly;
	  if (gcmp0(x))
	  {
	    z=cgetg(3,t_SER);
	    z[1]=evalvalp(valp(x)-valp(y))+evalvarn(vx);
	    return z;
	  }

	  z=cgetg(l,ty); p2=cgeti(l);
	  for (i=3; i<l; i++) p2[i]=!isexactzero((GEN)y[i]);
	  z[1]=evalvalp(valp(x)-valp(y))+evalvarn(vx);
	  z[2]=ldiv((GEN)x[2],(GEN)y[2]);
	  for (i=3; i<l; i++)
	  {
	    av=avma; p1=(GEN)x[i];
	    for (j=2; j<i; j++)
	      if (p2[i-j+2]) p1=gsub(p1,gmul((GEN)z[j],(GEN)y[i-j+2]));
	    tetpil=avma; p1=gdiv(p1,(GEN)y[2]);
	    z[i]=lpile(av,tetpil,p1);
	  }
	  normalize(&z); return z;

	case t_RFRAC: case t_RFRACN:
	  l=avma; p2=gmul(x,(GEN)y[2]); tetpil=avma;
	  return gerepile(l,tetpil,gdiv(p2,(GEN)y[1]));

	case t_VEC: case t_COL: case t_MAT:
	  err(gdiver12);
	default: err(typeer,"division");
      }

    case t_RFRAC: case t_RFRACN:
      switch(ty)
      {
	case t_POL:
	  l=avma; z=cgetg(lx,tx); z[2]=lmul((GEN)x[2],y);
	  if (tx==t_RFRAC)
	  { 
	    z[1]=x[1];
	    tetpil=avma; z=gerepile(l,tetpil,gred(z));
	  }
	  z[1]=lcopy((GEN)x[1]); return z;

	case t_SER:
	  l=avma; p2=gmul((GEN)x[2],y); tetpil=avma;
	  p2=gdiv((GEN)x[1],p2); return gerepile(l,tetpil,p2);

	case t_RFRAC: case t_RFRACN:
	  if (tx == t_RFRAC && ty == t_RFRAC) i=t_RFRAC; else i=t_RFRACN;
	  l=avma; z=cgetg(ly,i);
	  z[1]=lmul((GEN)x[1],(GEN)y[2]); z[2]=lmul((GEN)x[2],(GEN)y[1]);
	  if (i==t_RFRAC) { tetpil=avma; z=gerepile(l,tetpil,gred(z)); }
	  return z;

	case t_VEC: case t_COL: case t_MAT:
	  err(gdiver16);
	default: err(typeer,"division");
      }

    case t_QFR:
      l=signe(y[2]); setsigne(y[2],-l); k=signe(y[4]); setsigne(y[4],-k);
      z=compreal(x,y); setsigne(y[2],l); setsigne(y[4],k);
      return z;

    case t_QFI:
      l=signe(y[2]); setsigne(y[2],-l);
      z=compimag(x,y); setsigne(y[2],l);
      return z;

    case t_VEC: case t_COL: case t_MAT:
      if (!is_matvec_t(ty))
      {
	z=cgetg(lx,tx);
	for (i=1; i<lx; i++) z[i]=ldiv((GEN)x[i],y);
	return z;
      }

      if (ty==t_MAT && ly!=1 && lg(y[1])==ly)
      {
	l=avma;
	p1=invmat(y); tetpil=avma;
	return gerepile(l,tetpil,gmul(x,p1));
      }
      err(gdiver17);
  }
  err(typeer,"division");
  return NULL; /* not reached */
}
